/*
    Crystal Space Entity Layer
    Copyright (C) 2003 by Jorrit Tyberghein

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "cssysdef.h"
#include "csqsqrt.h"
#include "iutil/plugin.h"
#include "iutil/virtclk.h"
#include "iutil/eventq.h"
#include "iutil/evdefs.h"
#include "iutil/event.h"
#include "csutil/debug.h"
#include "csutil/objreg.h"
#include "csgeom/vector3.h"
#include "csgeom/math3d.h"
#include "iengine/mesh.h"
#include "iengine/movable.h"
#include "ivaria/reporter.h"
#include "ivaria/dynamics.h"

#include "plugins/propclass/dynmove/dynmove.h"
#include "physicallayer/pl.h"
#include "physicallayer/entity.h"
#include "physicallayer/persist.h"

//---------------------------------------------------------------------------

CS_IMPLEMENT_PLUGIN

CEL_IMPLEMENT_FACTORY(DynamicSystem, "pcdynsys")
CEL_IMPLEMENT_FACTORY(DynamicBody, "pcdynbody")

static void Report (iObjectRegistry* object_reg, const char* msg, ...)
{
  va_list arg;
  va_start (arg, msg);

  csRef<iReporter> rep (CS_QUERY_REGISTRY (object_reg, iReporter));
  if (rep)
    rep->ReportV (CS_REPORTER_SEVERITY_ERROR, "cel.persistence",
    	msg, arg);
  else
  {
    csPrintfV (msg, arg);
    csPrintf ("\n");
    fflush (stdout);
  }

  va_end (arg);
}

#ifdef NOTIFY_UNUSED
static void Notify (iObjectRegistry* object_reg, const char* msg, ...)
{
  va_list arg;
  va_start (arg, msg);

  csRef<iReporter> rep (CS_QUERY_REGISTRY (object_reg, iReporter));
  if (rep)
    rep->ReportV (CS_REPORTER_SEVERITY_NOTIFY, "cel.persistence",
    	msg, arg);
  else
  {
    csPrintfV (msg, arg);
    csPrintf ("\n");
    fflush (stdout);
  }

  va_end (arg);
}
#endif

//---------------------------------------------------------------------------

SCF_IMPLEMENT_IBASE_EXT (celPcDynamicSystem)
  SCF_IMPLEMENTS_EMBEDDED_INTERFACE (iPcDynamicSystem)
SCF_IMPLEMENT_IBASE_EXT_END

SCF_IMPLEMENT_EMBEDDED_IBASE (celPcDynamicSystem::PcDynamicSystem)
  SCF_IMPLEMENTS_INTERFACE (iPcDynamicSystem)
SCF_IMPLEMENT_EMBEDDED_IBASE_END

celPcDynamicSystem::celPcDynamicSystem (iObjectRegistry* object_reg)
	: celPcCommon (object_reg)
{
  SCF_CONSTRUCT_EMBEDDED_IBASE (scfiPcDynamicSystem);
  delta = 0.01f;
  vc = CS_QUERY_REGISTRY (object_reg, iVirtualClock);
  pl->CallbackEveryFrame ((iCelTimerListener*)this, cscmdPreProcess);
}

celPcDynamicSystem::~celPcDynamicSystem ()
{
  SCF_DESTRUCT_EMBEDDED_IBASE (scfiPcDynamicSystem);
}

void celPcDynamicSystem::ProcessForces (float dt)
{
// @@@ Note: the code below should really factor in the
// fact that 'dt' may be smaller than delta and also that the
// remaining time for a force may be smaller than the 'dt' value
// given here.
  size_t i;
  for (i = 0 ; i < forces.Length () ; i++)
  {
    celForce& f = forces[i];
    if (f.frame)
    {
      f.body->GetBody ()->AddForce (f.force);
    }
    else if (f.seconds > 0)
    {
      f.body->GetBody ()->AddForce (f.force);
      f.seconds -= dt;
      if (f.seconds <= 0)
      {
        forces.DeleteIndex (i);
	i--;
      }
    }
  }
}

void celPcDynamicSystem::TickEveryFrame ()
{
  GetDynamicSystem ();
  csTicks elapsed_time = vc->GetElapsedTicks ();
  float et = float (elapsed_time) / 1000.0;
  while (et >= delta)
  {
    ProcessForces (delta);
    dynamics->Step (delta);
    et -= delta;
  }
  // Do the remainder.
  if (et > SMALL_EPSILON)
  {
    ProcessForces (et);
    dynamics->Step (et);
  }

  // Delete all expired forces and forces that were only
  // meant to be here for one frame.
  size_t i;
  for (i = 0 ; i < forces.Length () ; i++)
  {
    celForce& f = forces[i];
    if (f.frame || f.seconds <= 0)
    {
      forces.DeleteIndex (i);
      i--;
    }
  }
}

iDynamicSystem* celPcDynamicSystem::GetDynamicSystem ()
{
  if (!dynsystem)
  {
    dynamics = CS_QUERY_REGISTRY (object_reg, iDynamics);
    if (!dynamics)
    {
      csRef<iPluginManager> plugmgr = CS_QUERY_REGISTRY (object_reg,
      	iPluginManager);
      dynamics = CS_LOAD_PLUGIN (plugmgr, "crystalspace.dynamics.ode",
      	iDynamics);
      if (dynamics)
      {
        object_reg->Register (dynamics, "iDynamics");
      }
    }
    if (!dynamics)
    {
      Report (object_reg, "Can't find dynamic subsystem!");
      return 0;
    }
    dynsystem = dynamics->CreateSystem ();
    dynsystem->SetGravity (csVector3 (0, -9.8f, 0));
  }
  return dynsystem;
}

#define DYNSYS_SERIAL 1

csPtr<iCelDataBuffer> celPcDynamicSystem::Save ()
{
  csRef<iCelDataBuffer> databuf = pl->CreateDataBuffer (DYNSYS_SERIAL);
  databuf->Add (dynsystem->GetGravity ());
  databuf->Add (dynsystem->GetLinearDampener ());
  databuf->Add (dynsystem->GetRollingDampener ());
  databuf->Add (delta);
  return csPtr<iCelDataBuffer> (databuf);
}

bool celPcDynamicSystem::Load (iCelDataBuffer* databuf)
{
  int serialnr = databuf->GetSerialNumber ();
  if (serialnr != DYNSYS_SERIAL)
  {
    Report (object_reg, "serialnr != DYNSYS_SERIAL.  Cannot load.");
    return false;
  }

  csVector3 gravity;
  databuf->GetVector3 (gravity);
  float lineardampener = databuf->GetFloat ();
  float rollingdampener = databuf->GetFloat ();
  delta = databuf->GetFloat ();

  if (!GetDynamicSystem ())
    return false;

  dynsystem->SetGravity (gravity);
  dynsystem->SetLinearDampener (lineardampener);
  dynsystem->SetRollingDampener (rollingdampener);

  return true;
}

void celPcDynamicSystem::AddForceDuration (iPcDynamicBody* body,
  	const csVector3& force, float seconds)
{
  celForce f;
  f.body = body;
  f.seconds = seconds;
  f.frame = false;
  f.force = force;
  forces.Push (f);
}

void celPcDynamicSystem::AddForceFrame (iPcDynamicBody* body,
	const csVector3& force)
{
  celForce f;
  f.body = body;
  f.seconds = 0;
  f.frame = true;
  f.force = force;
  forces.Push (f);
}

void celPcDynamicSystem::ClearForces (iPcDynamicBody* body)
{
  size_t i;
  for (i = 0 ; i < forces.Length () ; i++)
  {
    if (forces[i].body == body)
    {
      forces.DeleteIndex (i);
      i--;
    }
  }
}

void celPcDynamicSystem::ClearAllForces ()
{
  forces.DeleteAll ();
}

//---------------------------------------------------------------------------

SCF_IMPLEMENT_IBASE_EXT (celPcDynamicBody)
  SCF_IMPLEMENTS_EMBEDDED_INTERFACE (iPcDynamicBody)
SCF_IMPLEMENT_IBASE_EXT_END

SCF_IMPLEMENT_EMBEDDED_IBASE (celPcDynamicBody::PcDynamicBody)
  SCF_IMPLEMENTS_INTERFACE (iPcDynamicBody)
SCF_IMPLEMENT_EMBEDDED_IBASE_END

celPcDynamicBody::celPcDynamicBody (iObjectRegistry* object_reg)
	: celPcCommon (object_reg)
{
  SCF_CONSTRUCT_EMBEDDED_IBASE (scfiPcDynamicBody);
  DG_TYPE (this, "celPcDynamicBody()");

  btype = CEL_BODY_INVALID;
  bdata = 0;

  friction = 1;
  elasticity = 0;
  softness = 0;
  density = 1;
  mass = 1;

  is_static = false;
}

celPcDynamicBody::~celPcDynamicBody ()
{
  delete bdata;
  if (dynsystem)
    dynsystem->ClearForces (&scfiPcDynamicBody);
  SCF_DESTRUCT_EMBEDDED_IBASE (scfiPcDynamicBody);
}

#define DYNBODY_SERIAL 1

csPtr<iCelDataBuffer> celPcDynamicBody::Save ()
{
  csRef<iCelDataBuffer> databuf = pl->CreateDataBuffer (DYNBODY_SERIAL);
  int body_info_cnt = 0;
  switch (btype)
  {
    case CEL_BODY_SPHERE: body_info_cnt = 2; break;
    case CEL_BODY_BOX: body_info_cnt = 5; break;
    case CEL_BODY_CYLINDER: body_info_cnt = 2; break;
    case CEL_BODY_PLANE: body_info_cnt = 4; break;
    case CEL_BODY_MESH: body_info_cnt = 0; break;
  }
  csRef<iCelPropertyClass> pc;
  if (pcmesh) pc = SCF_QUERY_INTERFACE (pcmesh, iCelPropertyClass);
  databuf->Add (pc);
  if (dynsystem) pc = SCF_QUERY_INTERFACE (dynsystem, iCelPropertyClass);
  databuf->Add (pc);
  databuf->Add (int32 (btype));
  switch (btype)
  {
    case CEL_BODY_SPHERE:
      {
        sphere_data* sd = (sphere_data*)bdata;
        databuf->Add (sd->radius);
        databuf->Add (sd->offset);
      }
      break;
    case CEL_BODY_BOX:
      {
        box_data* bd = (box_data*)bdata;
        databuf->Add (bd->size);
        const csMatrix3& m = bd->transform.GetO2T ();
        databuf->Add (m.Row1 ());
        databuf->Add (m.Row2 ());
        databuf->Add (m.Row3 ());
        databuf->Add (bd->transform.GetO2TTranslation ());
      }
      break;
    case CEL_BODY_CYLINDER:
      {
        cylinder_data* cd = (cylinder_data*)bdata;
        databuf->Add (cd->length);
        databuf->Add (cd->radius);
        const csMatrix3& m = cd->transform.GetO2T ();
        databuf->Add (m.Row1 ());
        databuf->Add (m.Row2 ());
        databuf->Add (m.Row3 ());
        databuf->Add (cd->transform.GetO2TTranslation ());
      }
      break;
    case CEL_BODY_PLANE:
      {
        plane_data* pd = (plane_data*)bdata;
        databuf->Add (pd->plane.A ());
        databuf->Add (pd->plane.B ());
        databuf->Add (pd->plane.C ());
        databuf->Add (pd->plane.D ());
      }
      break;
    case CEL_BODY_MESH:
      break;
  }
  databuf->Add (friction);
  databuf->Add (elasticity);
  databuf->Add (softness);
  databuf->Add (density);
  databuf->Add (mass);
  databuf->Add (is_static);

  return csPtr<iCelDataBuffer> (databuf);
}

bool celPcDynamicBody::Load (iCelDataBuffer* databuf)
{
  int serialnr = databuf->GetSerialNumber ();
  if (serialnr != DYNBODY_SERIAL)
  {
    Report (object_reg, "serialnr != DYNBODY_SERIAL.  Cannot load.");
    return false;
  }
// @@@ TODO
#if 0
  celData* cd;
  collider_wrap = 0;
  no_collider = false;
  cd = databuf->GetData (0);
  if (!cd)
  {
    Report (object_reg,"iPcMesh prop class not specified.  Cannot load.");
    return false;
  }
  csRef<iPcMesh> pcm;
  if (cd->value.pc)
    pcm = SCF_QUERY_INTERFACE (cd->value.pc, iPcMesh);
  SetMesh (pcm);
#endif
  return true;
}

iRigidBody* celPcDynamicBody::GetBody ()
{
  if (!body)
  {
    body = dynsystem->GetDynamicSystem ()->CreateBody ();
  }
  return body;
}

iPcMesh* celPcDynamicBody::GetMesh ()
{
  if (!pcmesh)
  {
    pcmesh = CEL_QUERY_PROPCLASS (entity->GetPropertyClassList (), iPcMesh);
  }
  return pcmesh;
}

void celPcDynamicBody::MakeStatic (bool stat)
{
  if (stat == is_static) return;
  is_static = stat;
  GetBody ();
  if (is_static)
  {
    body->SetMoveCallback (0);
    body->MakeStatic ();
  }
  else
  {
    // Restore move callback???@@@
    body->MakeDynamic ();
  }
}

void celPcDynamicBody::AttachColliderSphere (float radius,
	const csVector3& offset)
{
  GetBody ();
  GetMesh ();
  body->AttachColliderSphere (radius, offset, friction, density,
  	elasticity, softness);
  body->AdjustTotalMass (mass);
  if (pcmesh)
  {
    body->SetTransform (pcmesh->GetMesh ()->GetMovable ()->GetFullTransform ());
    body->AttachMesh (pcmesh->GetMesh ());
  }

  delete bdata;
  bdata = new sphere_data (radius, offset);
  btype = CEL_BODY_SPHERE;
}

void celPcDynamicBody::AttachColliderCylinder (float length, float radius,
	const csOrthoTransform& trans)
{
  GetBody ();
  GetMesh ();
  body->AttachColliderCylinder (length, radius, trans, friction, density,
  	elasticity, softness);
  body->AdjustTotalMass (mass);
  if (pcmesh)
  {
    body->SetTransform (pcmesh->GetMesh ()->GetMovable ()->GetFullTransform ());
    body->AttachMesh (pcmesh->GetMesh ());
  }

  delete bdata;
  bdata = new cylinder_data (length, radius, trans);
  btype = CEL_BODY_CYLINDER;
}

void celPcDynamicBody::AttachColliderBox (const csVector3& size,
  	const csOrthoTransform& trans)
{
  GetBody ();
  GetMesh ();
  body->AttachColliderBox (size, trans, friction, density, elasticity,
  	softness);
  body->AdjustTotalMass (mass);
  if (pcmesh)
  {
    CS_ASSERT (pcmesh->GetMesh () != 0);
    body->SetTransform (pcmesh->GetMesh ()->GetMovable ()->GetFullTransform ());
    body->AttachMesh (pcmesh->GetMesh ());
  }

  delete bdata;
  bdata = new box_data (size, trans);
  btype = CEL_BODY_BOX;
}

void celPcDynamicBody::AttachColliderPlane (const csPlane3& plane)
{
  GetBody ();
  GetMesh ();
  body->AttachColliderPlane (plane, friction, density, elasticity,
  	softness);
  body->AdjustTotalMass (mass);
  if (pcmesh)
  {
    body->SetTransform (pcmesh->GetMesh ()->GetMovable ()->GetFullTransform ());
    body->AttachMesh (pcmesh->GetMesh ());
  }

  delete bdata;
  bdata = new plane_data (plane);
  btype = CEL_BODY_PLANE;
}

void celPcDynamicBody::AttachColliderMesh ()
{
  GetBody ();
  GetMesh ();
  iMeshWrapper* mesh = pcmesh->GetMesh ();
  body->AttachColliderMesh (mesh, mesh->GetMovable ()->GetFullTransform (),
  	friction, density, elasticity, softness);
  body->AdjustTotalMass (mass);
  body->SetTransform (mesh->GetMovable ()->GetFullTransform ());
  body->AttachMesh (mesh);

  delete bdata;
  bdata = 0;
  btype = CEL_BODY_MESH;
}

void celPcDynamicBody::AddForceOnce (const csVector3& force)
{
  GetBody ();
  body->AddForce (force);
}

void celPcDynamicBody::AddForceDuration (const csVector3& force, float seconds)
{
  dynsystem->AddForceDuration (&scfiPcDynamicBody, force, seconds);
}

void celPcDynamicBody::AddForceFrame (const csVector3& force)
{
  dynsystem->AddForceFrame (&scfiPcDynamicBody, force);
}

void celPcDynamicBody::ClearForces ()
{
  dynsystem->ClearForces (&scfiPcDynamicBody);
}

